@@ -1,19 +1,20 @@
 """Django database model for SatNOGS DB"""
-from __future__ import absolute_import, division, print_function, \
-    unicode_literals
-
 import logging
 from os import path
 from uuid import uuid4
 
+import h5py
+import satnogsdecoders
 from django.conf import settings
 from django.contrib.auth.models import User
-from django.core.validators import MaxValueValidator, MinValueValidator
+from django.core.exceptions import ValidationError
+from django.core.validators import MaxValueValidator, MinValueValidator, \
+    URLValidator
 from django.db import models
 from django.db.models import OuterRef, Subquery
 from django.db.models.signals import post_save, pre_save
-from django.utils.encoding import python_2_unicode_compatible
 from django.utils.timezone import now
+from django_countries.fields import CountryField
 from markdown import markdown
 from shortuuidfield import ShortUUIDField
 
@@ -30,6 +31,11 @@ SERVICE_TYPE = [
     'Maritime', 'Meteorological', 'Mobile', 'Radiolocation', 'Radionavigational',
     'Space Operation', 'Space Research', 'Standard Frequency and Time Signal', 'Unknown'
 ]
+URL_REGEX = r"(?:http(s)?:\/\/)?[\w.-]+(?:\.[\w\.-]+)+[\w\-\._~:/?#[\]@!\$&'\(\)\*\+,;=.]+$"
+MIN_FREQ = 0
+MAX_FREQ = 40000000000
+MIN_FREQ_MSG = "Ensure this value is greater than or equal to 0Hz"
+MAX_FREQ_MSG = "Ensure this value is less than or equal to 40Ghz"
 
 
 def _name_exported_frames(instance, filename):  # pylint: disable=W0613
@@ -67,7 +73,18 @@ def _set_is_decoded(sender, instance, **kwargs):  # pylint: disable=W0613
     instance.is_decoded = instance.payload_decoded != ''
 
 
-@python_2_unicode_compatible
+def _extract_network_obs_id(sender, instance, created, **kwargs):  # pylint: disable=W0613
+    post_save.disconnect(_extract_network_obs_id, sender=Artifact)
+    try:
+        with h5py.File(instance.artifact_file, 'r') as h5_file:
+            instance.network_obs_id = h5_file.attrs["observation_id"]
+    except OSError as error:
+        LOGGER.warning(error)
+
+    instance.save()
+    post_save.connect(_extract_network_obs_id, sender=Artifact)
+
+
 class Mode(models.Model):
     """A satellite transmitter RF mode. For example: FM"""
     name = models.CharField(max_length=25, unique=True)
@@ -76,14 +93,31 @@ class Mode(models.Model):
         return self.name
 
 
-@python_2_unicode_compatible
+class Operator(models.Model):
+    """Satellite Owner/Operator"""
+    name = models.CharField(max_length=255, unique=True)
+    names = models.TextField(blank=True)
+    description = models.TextField(blank=True)
+    website = models.URLField(
+        blank=True, validators=[URLValidator(schemes=['http', 'https'], regex=URL_REGEX)]
+    )
+
+    def __str__(self):
+        return self.name
+
+
 class Satellite(models.Model):
     """Model for all the satellites."""
     norad_cat_id = models.PositiveIntegerField()
     name = models.CharField(max_length=45)
     names = models.TextField(blank=True)
     description = models.TextField(blank=True)
-    dashboard_url = models.URLField(blank=True, null=True, max_length=200)
+    dashboard_url = models.URLField(
+        blank=True,
+        null=True,
+        max_length=200,
+        validators=[URLValidator(schemes=['http', 'https'], regex=URL_REGEX)]
+    )
     image = models.ImageField(upload_to='satellites', blank=True, help_text='Ideally: 250x250')
     tle1 = models.CharField(max_length=200, blank=True)
     tle2 = models.CharField(max_length=200, blank=True)
@@ -93,6 +127,22 @@ class Satellite(models.Model):
     )
     decayed = models.DateTimeField(null=True, blank=True)
 
+    # new fields below, metasat etc
+    # countries is multiple for edge cases like ISS/Zarya
+    countries = CountryField(blank=True, multiple=True, blank_label='(select countries)')
+    website = models.URLField(
+        blank=True, validators=[URLValidator(schemes=['http', 'https'], regex=URL_REGEX)]
+    )
+    launched = models.DateTimeField(null=True, blank=True)
+    deployed = models.DateTimeField(null=True, blank=True)
+    operator = models.ForeignKey(
+        Operator,
+        blank=True,
+        null=True,
+        on_delete=models.SET_NULL,
+        related_name='satellite_operator'
+    )
+
     class Meta:
         ordering = ['norad_cat_id']
 
@@ -150,6 +200,15 @@ class Satellite(models.Model):
         decoder_count = Telemetry.objects.filter(satellite=self.id).exclude(decoder='').count()
         return decoder_count
 
+    @property
+    def transmitter_count(self):
+        """Returns number of approved transmitter suggestions for this Satellite
+
+        :returns: number of approved transmitter suggestions for this Satellite
+        """
+        approved_count = Transmitter.objects.filter(satellite=self.id).count()
+        return approved_count
+
     @property
     def tle_redistributable(self):
         """Returns True if re-distribution of the TLE is allowed, False otherwise
@@ -172,55 +231,162 @@ class Satellite(models.Model):
             'redistributable': self.tle_redistributable
         }
 
+    @property
+    def latest_data(self):
+        """Returns the latest DemodData for this Satellite
+
+        :returns: dict with most recent DemodData for this Satellite
+        """
+        data = DemodData.objects.filter(satellite=self.id).order_by('-id')[:1]
+        return {
+            'data_id': data[0].data_id,
+            'payload_frame': data[0].payload_frame,
+            'timestamp': data[0].timestamp,
+            'is_decoded': data[0].is_decoded,
+            'station': data[0].station,
+            'observer': data[0].observer,
+        }
+
+    @property
+    def needs_help(self):
+        """Returns a boolean based on whether or not this Satellite could
+            use some editorial help based on a configurable threshold
+
+        :returns: bool
+        """
+        score = 0
+        if self.description and self.description != '':
+            score += 1
+        if self.countries and self.countries != '':
+            score += 1
+        if self.website and self.website != '':
+            score += 1
+        if self.names and self.names != '':
+            score += 1
+        if self.launched and self.launched != '':
+            score += 1
+        if self.operator and self.operator != '':
+            score += 1
+        if self.image and self.image != '':
+            score += 1
+
+        return score <= 2
+
     def __str__(self):
         return '{0} - {1}'.format(self.norad_cat_id, self.name)
 
 
-@python_2_unicode_compatible
 class TransmitterEntry(models.Model):
     """Model for satellite transmitters."""
     uuid = ShortUUIDField(db_index=True)
-    description = models.TextField()
+    description = models.TextField(
+        help_text='Short description for this entry, like: UHF 9k6 AFSK Telemetry'
+    )
     status = models.CharField(
-        choices=list(zip(TRANSMITTER_STATUS, TRANSMITTER_STATUS)), max_length=8, default='active'
+        choices=list(zip(TRANSMITTER_STATUS, TRANSMITTER_STATUS)),
+        max_length=8,
+        default='active',
+        help_text='Functional state of this transmitter'
     )
     type = models.CharField(
         choices=list(zip(TRANSMITTER_TYPE, TRANSMITTER_TYPE)),
         max_length=11,
         default='Transmitter'
     )
-    uplink_low = models.BigIntegerField(blank=True, null=True)
-    uplink_high = models.BigIntegerField(blank=True, null=True)
-    uplink_drift = models.IntegerField(blank=True, null=True)
-    downlink_low = models.BigIntegerField(blank=True, null=True)
-    downlink_high = models.BigIntegerField(blank=True, null=True)
-    downlink_drift = models.IntegerField(blank=True, null=True)
+    uplink_low = models.BigIntegerField(
+        blank=True,
+        null=True,
+        validators=[
+            MinValueValidator(MIN_FREQ, message=MIN_FREQ_MSG),
+            MaxValueValidator(MAX_FREQ, message=MAX_FREQ_MSG)
+        ],
+        help_text='Frequency (in Hz) for the uplink, or bottom of the uplink range for a \
+            transponder'
+    )
+    uplink_high = models.BigIntegerField(
+        blank=True,
+        null=True,
+        validators=[
+            MinValueValidator(MIN_FREQ, message=MIN_FREQ_MSG),
+            MaxValueValidator(MAX_FREQ, message=MAX_FREQ_MSG)
+        ],
+        help_text='Frequency (in Hz) for the top of the uplink range for a transponder'
+    )
+    uplink_drift = models.IntegerField(
+        blank=True,
+        null=True,
+        validators=[MinValueValidator(-99999), MaxValueValidator(99999)],
+        help_text='Receiver drift from the published uplink frequency, stored in parts \
+            per billion (PPB)'
+    )
+    downlink_low = models.BigIntegerField(
+        blank=True,
+        null=True,
+        validators=[
+            MinValueValidator(MIN_FREQ, message=MIN_FREQ_MSG),
+            MaxValueValidator(MAX_FREQ, message=MAX_FREQ_MSG)
+        ],
+        help_text='Frequency (in Hz) for the downlink, or bottom of the downlink range \
+            for a transponder'
+    )
+    downlink_high = models.BigIntegerField(
+        blank=True,
+        null=True,
+        validators=[
+            MinValueValidator(MIN_FREQ, message=MIN_FREQ_MSG),
+            MaxValueValidator(MAX_FREQ, message=MAX_FREQ_MSG)
+        ],
+        help_text='Frequency (in Hz) for the top of the downlink range for a transponder'
+    )
+    downlink_drift = models.IntegerField(
+        blank=True,
+        null=True,
+        validators=[MinValueValidator(-99999), MaxValueValidator(99999)],
+        help_text='Transmitter drift from the published downlink frequency, stored in \
+            parts per billion (PPB)'
+    )
     downlink_mode = models.ForeignKey(
         Mode,
         blank=True,
         null=True,
         on_delete=models.SET_NULL,
-        related_name='transmitter_downlink_entries'
+        related_name='transmitter_downlink_entries',
+        help_text='Modulation mode for the downlink'
     )
     uplink_mode = models.ForeignKey(
         Mode,
         blank=True,
         null=True,
         on_delete=models.SET_NULL,
-        related_name='transmitter_uplink_entries'
+        related_name='transmitter_uplink_entries',
+        help_text='Modulation mode for the uplink'
+    )
+    invert = models.BooleanField(
+        default=False, help_text='True if this is an inverted transponder'
+    )
+    baud = models.FloatField(
+        validators=[MinValueValidator(0)],
+        blank=True,
+        null=True,
+        help_text='The number of modulated symbols that the transmitter sends every second'
     )
-    invert = models.BooleanField(default=False)
-    baud = models.FloatField(validators=[MinValueValidator(0)], blank=True, null=True)
     satellite = models.ForeignKey(
         Satellite, null=True, related_name='transmitter_entries', on_delete=models.SET_NULL
     )
     reviewed = models.BooleanField(default=False)
     approved = models.BooleanField(default=False)
-    created = models.DateTimeField(default=now)
-    citation = models.CharField(max_length=512, default='CITATION NEEDED - https://xkcd.com/285/')
+    created = models.DateTimeField(default=now, help_text='Timestamp for this entry or edit')
+    citation = models.CharField(
+        max_length=512,
+        default='CITATION NEEDED - https://xkcd.com/285/',
+        help_text='A reference (preferrably URL) for this entry or edit'
+    )
     user = models.ForeignKey(User, null=True, on_delete=models.SET_NULL)
     service = models.CharField(
-        choices=zip(SERVICE_TYPE, SERVICE_TYPE), max_length=34, default='Unknown'
+        choices=zip(SERVICE_TYPE, SERVICE_TYPE),
+        max_length=34,
+        default='Unknown',
+        help_text='The published usage category for this transmitter'
     )
 
     # NOTE: future fields will need to be added to forms.py and to
@@ -240,6 +406,36 @@ class TransmitterEntry(models.Model):
         self.id = None  # pylint: disable=C0103, W0201
         super(TransmitterEntry, self).save()
 
+    def clean(self):
+        if self.type == TRANSMITTER_TYPE[0]:
+            if self.uplink_low is not None or self.uplink_high is not None \
+                    or self.uplink_drift is not None:
+                raise ValidationError("Uplink shouldn't be filled in for a transmitter")
+
+            if self.downlink_high:
+                raise ValidationError(
+                    "Downlink high frequency shouldn't be filled in for a transmitter"
+                )
+
+        elif self.type == TRANSMITTER_TYPE[1]:
+            if self.uplink_high is not None or self.downlink_high is not None:
+                raise ValidationError("Frequency range shouldn't be filled in for a transceiver")
+
+        elif self.type == TRANSMITTER_TYPE[2]:
+            if self.downlink_low is not None and self.downlink_high is not None:
+                if self.downlink_low > self.downlink_high:
+                    raise ValidationError(
+                        "Downlink low frequency must be lower or equal \
+                        than downlink high frequency"
+                    )
+
+            if self.uplink_low is not None and self.uplink_high is not None:
+                if self.uplink_low > self.uplink_high:
+                    raise ValidationError(
+                        "Uplink low frequency must be lower or equal \
+                        than uplink high frequency"
+                    )
+
 
 class TransmitterSuggestionManager(models.Manager):  # pylint: disable=R0903
     """Django Manager for TransmitterSuggestions
@@ -252,7 +448,6 @@ class TransmitterSuggestionManager(models.Manager):  # pylint: disable=R0903
         return TransmitterEntry.objects.filter(reviewed=False)
 
 
-@python_2_unicode_compatible
 class TransmitterSuggestion(TransmitterEntry):
     """TransmitterSuggestion is an unreviewed TransmitterEntry object"""
     objects = TransmitterSuggestionManager()
@@ -282,7 +477,6 @@ class TransmitterManager(models.Manager):  # pylint: disable=R0903
         ).filter(created=Subquery(subquery.values('created')[:1]))
 
 
-@python_2_unicode_compatible
 class Transmitter(TransmitterEntry):
     """Associates a generic Transmitter object with their TransmitterEntries
     that are managed by TransmitterManager
@@ -296,7 +490,6 @@ class Transmitter(TransmitterEntry):
         proxy = True
 
 
-@python_2_unicode_compatible
 class Telemetry(models.Model):
     """Model for satellite telemetry decoders."""
     satellite = models.ForeignKey(
@@ -313,8 +506,19 @@ class Telemetry(models.Model):
     def __str__(self):
         return self.name
 
+    def get_kaitai_fields(self):
+        """Return an empty-value dict of fields for this kaitai.io struct
+        Beware the overuse of "decoder" in satnogsdecoders and "decoder" the
+        field above in this Telemetry model"""
+        results = {}
+        try:
+            decoder_class = getattr(satnogsdecoders.decoder, self.decoder.capitalize())
+            results = satnogsdecoders.decoder.get_fields(decoder_class, empty=True)
+        except AttributeError:
+            pass
+        return results
+
 
-@python_2_unicode_compatible
 class DemodData(models.Model):
     """Model for satellite for observation data."""
     satellite = models.ForeignKey(
@@ -368,7 +572,6 @@ post_save.connect(_gen_observer, sender=DemodData)
 pre_save.connect(_set_is_decoded, sender=DemodData)
 
 
-@python_2_unicode_compatible
 class ExportedFrameset(models.Model):
     """Model for exported frames."""
     created = models.DateTimeField(auto_now_add=True)
@@ -377,3 +580,17 @@ class ExportedFrameset(models.Model):
     exported_file = models.FileField(upload_to=_name_exported_frames, blank=True, null=True)
     start = models.DateTimeField(blank=True, null=True)
     end = models.DateTimeField(blank=True, null=True)
+
+
+class Artifact(models.Model):
+    """Model for observation artifacts."""
+
+    artifact_file = models.FileField(upload_to='artifacts/', blank=True, null=True)
+
+    network_obs_id = models.BigIntegerField(blank=True, null=True)
+
+    def __str__(self):
+        return 'artifact-{0}'.format(self.id)
+
+
+post_save.connect(_extract_network_obs_id, sender=Artifact)
